/**
 * Diese Tensorflow Libraries herunterladen und in BlueJ unter Einstellungen
 * im Reiter <Bibliotheken> hinzufügen. BlueJ danach neustarten.
 * Es sollte jeweils die aktuelle Version gewählt werden. Bei der Konzeption 
 * dieser Klasse waren die u.g. Versionen aktuell.
 * 
 * 1. TensorFlow Core API Library: org.tensorflow -> tensorflow-core-api
 *      https://mvnrepository.com/artifact/org.tensorflow/tensorflow-core-api
 *          -> tensorflow-core-api-0.4.0.jar
 *      
 * 2.   bei 1. zusätzlich mit Klick auf "View All" folgende Seite öffnen
 *      https://repo1.maven.org/maven2/org/tensorflow/tensorflow-core-api/0.4.0/
 *      Dort die entsprechende Library für das Betriebssystem herunterladen, z.B.:
 *          -> tensorflow-core-api-0.4.0-macosx-x86_64.jar
 *          -> tensorflow-core-api-0.4.0-windows-x86_64.jar
 *          -> tensorflow-core-api-0.4.0-linux-x86_64.jar 
 *      
 * 3. TensorFlow Framework Library:  org.tensorflow -> tensorflow-framework
 *      https://mvnrepository.com/artifact/org.tensorflow/tensorflow-framework/0.4.0
 *          -> tensorflow-framework-0.4.0.jar      
 *          
 * 4. Protocol Buffers [Core]: com.google.protobuf -> protobuf-java
 *      https://mvnrepository.com/artifact/com.google.protobuf/protobuf-java
 *          -> protobuf-java-4.0.0-rc-2.jar
 * 
 * 5. JavaCPP: org.bytedeco -> javacpp
 *      https://mvnrepository.com/artifact/org.bytedeco/javacpp
 *          -> javacpp-1.5.7.jar
 * 
 * 6. TensorFlow NdArray Library:  org.tensorflow -> ndarray
 *      https://mvnrepository.com/artifact/org.tensorflow/ndarray
 *          -> ndarray-0.3.3.jar
 */

import org.tensorflow.SavedModelBundle;
import org.tensorflow.Tensor;
import org.tensorflow.ndarray.FloatNdArray;
import org.tensorflow.ndarray.IntNdArray;
import org.tensorflow.ndarray.NdArrays;
import org.tensorflow.ndarray.Shape;
import org.tensorflow.types.TFloat32;
import org.tensorflow.types.TInt32;
import java.util.HashMap;
import java.util.Map;
import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;
import java.awt.Color;
import java.io.File;

public class TFImageModel {

    private SavedModelBundle model;
    private String inputLayerName;
    private String outputLayerName;
    private String keyName;
    private BufferedImage image;
    private float[] output;
    private String errorMessage;   
    private int width;
    private int height;
    private String dtype;
    private String[] labels;

    /**
     * Konstruktor, um ein Tensorflow Model in Java aus einem Ordner 
     * mit einem 'Saved Model' zu erstellen
     * 
     * @param Name des Ordners in dem das Tensorflow Model liegt. Der Ordner enthält
     *        üblicherweise eine Datei namens saved_model.pb und andere Unterordner
     *        mit Variablen, manchmal auch Assets. 
     *        Achtung: Gültige Pfadangaben sind unter Windows z.B. "C:\\models\\myModel"
     *        oder unter Mac OS/Linux z.B. "/Users/myname/myModel"
     *                      
     * @param Breite des Bildes, das klassifiziert werden soll
     * @param Höhe des Bildes, das klassifiziert werden soll
     *        Beide Werte können der Dokumentation des Tensorflow Models entnommen werden
     *        oder mit der Methode getModelInformation() hier abgefragt werden
     */
    public TFImageModel(String pFoldername, int pImageHeight, int pImageWidth) {
        width = pImageWidth;
        height = pImageHeight;
        errorMessage = "";        
        try {            
            model = SavedModelBundle.load(pFoldername, "serve");
            inputLayerName = model.signatures().get(0).getInputs().keySet().toString();
            outputLayerName = model.signatures().get(0).getOutputs().keySet().toString();
            inputLayerName = inputLayerName.substring(1, inputLayerName.length()-1);
            outputLayerName = outputLayerName.substring(1, outputLayerName.length()-1);
            keyName = model.signatures().get(0).key();

            dtype = "FLOAT";
            String s = model.signatures().get(0).toString();        
            if (s.substring(0, s.lastIndexOf("dtype")).contains("INT")) {
                dtype = "INT";
            }
        }
        catch (Exception e) {
            errorMessage += "\n" + e.toString();
            model = null;
            inputLayerName = null;
            outputLayerName = null;
            keyName = null;
            dtype = null;
            output = null;
            labels = null;
        }                
    }

    /**
     * Methode, die jederzeit benutzt werden kann, um zu sehen ob zum Zeitpunkt 
     * des Aufrufs bereits irgendwelche Fehler aufgetreten sind 
     * 
     * @return Zeichenkette mit allen potentiell aufgetretenen Fehlern seit der 
     * letzten Instanziierung
     */
    public String getErrorMessage() {
        return errorMessage;
    }

    /**
     * Die Methode liefert Informationen über das Model, normalerweise inkl.
     * layer names, image width & height, data types     
     * 
     * @return Zeichenkette mit Informationen über das geladenene Model
     */
    public String getModelInformation() {
        String infos = "";
        if (model != null) {
            try {
                for (int i=0; i<model.signatures().size(); i++) {
                    infos += model.signatures().get(i).toString();
                } 
            } 
            catch (Exception e) {
                errorMessage += "\n" + e.toString();
            }
            return infos;
        }
        else {
            return null;
        }
    }    

    /**
     * Mit dieser Methode kann ein Bild als Eingabe in das Model geladen werden
     * Im Anschluss kann dieses mit run() klassiziert werden.
     * 
     * @param Dateiname inkl. Pfad des Bildes
     *        Achtung: Gültige Pfadangaben sind unter Windows z.B. "C:\\images\\myImage.png"
     *        oder unter Mac OS / Linux z.B. "/Users/myname/myImage.png"
     */
    public void setData(String pFilename) {
        image = null;
        try {
            image = ImageIO.read(new File(pFilename));
            if (image.getWidth() != width || image.getHeight() != height) {
                errorMessage += "\nWrong image dimension.";
                image = null;
            }
        } 
        catch (Exception e) {
            errorMessage += "\n" + e.toString();
        }
    }

    /**
     * Methode zum Ausführen der Klassifikation.
     * Das Ergebnis der Klassifikation kann mit getOutputArray() oder getOutputLabel()
     * abgefragt werden.
     */
    public void run() {
        if (model != null && image != null && inputLayerName != null && outputLayerName != null && keyName != null && dtype != null) {
            try {
                Map<String, Tensor> feed_dict = null;
                if (dtype.equals("INT")) {
                    IntNdArray input_matrix = NdArrays.ofInts(Shape.of(1, height, width, 3));
                    input_matrix = fillIntNdArray(input_matrix, image);            
                    Tensor input_tensor = TInt32.tensorOf(input_matrix);
                    feed_dict = new HashMap<>();
                    feed_dict.put(inputLayerName, input_tensor); 

                }
                else { //FLOAT            
                    FloatNdArray input_matrix = NdArrays.ofFloats(Shape.of(1, height, width, 3));
                    input_matrix = fillFloatNdArray(input_matrix, image);            
                    Tensor input_tensor = TFloat32.tensorOf(input_matrix);
                    feed_dict = new HashMap<>();
                    feed_dict.put(inputLayerName, input_tensor);    
                }

                Map<String, Tensor> res = model.function(keyName).call(feed_dict);                
                Tensor output_tensor = res.get(outputLayerName);    
                output = new float[(int)(output_tensor.asRawTensor().data().asFloats().size())];
                for (int i=0; i<output.length; i++) {
                    output[i] = output_tensor.asRawTensor().data().asFloats().getFloat(i);
                }
            }
            catch (Exception e) {
                errorMessage += "\n" + e.toString();
            } 
        }
    }

    /**
     * Eine Methode, um ein Bild in ein Tensorflow-lesbares Format umzuwandeln
     */
    private FloatNdArray fillFloatNdArray(FloatNdArray pMatrix, BufferedImage pImage) {
        try {
            int w = pImage.getWidth();
            int h = pImage.getHeight();

            for (int i = 0; i < h; i++) {
                for (int j = 0; j < w; j++) {                 
                    Color mycolor = new Color(pImage.getRGB(j, i));
                    int red = mycolor.getRed();                    
                    int green = mycolor.getGreen();                    
                    int blue = mycolor.getBlue();
                    pMatrix.setFloat(red/255.0f, 0, i, j, 0);
                    pMatrix.setFloat(green/255.0f, 0, i, j, 1);
                    pMatrix.setFloat(blue/255.0f, 0, i, j, 2);
                }
            }            
        }
        catch (Exception e) {
            errorMessage += "\n" + e.toString();
        }
        return pMatrix;
    }

    /**
     * Eine Methode, um ein Bild in ein Tensorflow-lesbares Format umzuwandeln
     */
    private IntNdArray fillIntNdArray(IntNdArray pMatrix, BufferedImage pImage) {        
        try {
            int w = pImage.getWidth();
            int h = pImage.getHeight();                

            for (int i = 0; i < h; i++) {
                for (int j = 0; j < w; j++) {                 
                    Color mycolor = new Color(pImage.getRGB(j, i));
                    int red = mycolor.getRed();
                    int green = mycolor.getGreen();
                    int blue = mycolor.getBlue();
                    pMatrix.setInt(red, 0, i, j, 0);
                    pMatrix.setInt(green, 0, i, j, 1);
                    pMatrix.setInt(blue, 0, i, j, 2);                                       
                }
            }
        }
        catch (Exception e) {
            errorMessage += "\n" + e.toString();            
        }
        return pMatrix;        
    }

    /**
     * Diese Methode ist eine von zwei Möglichkeiten, das Ergebnis einer Klassifikation
     * zu erhalten nachdem ein Bild mit run() klassifiziert wurde     
     * 
     * @return Array mit gleich vielen Feldern wie Ausgabeneuronen. Je höher der Wert,
     * desto wahrscheinlicher ist es, dass das Bild zur Klasse gehört, die das Ausgabe-
     * neuron repräsentiert.     
     */
    public float[] getOutputArray() {
        return output;
    }

    /**
     * Mit der Methode kann ein String-Array übergeben werden, das die Klassenlabels enthält.
     * So kann jedem Ausgabeneuron genau ein Text zugeordnet werden und eine Ausgabe wie
     * "Das war eine Katze" kann damit relaisiert werden.
     * 
     * @param: Output Labels (Klassen) als String-Array
     *         Die Klassen müssen in der richtigen Reihenfolge und passend zum Tensorflow
     *         Model sein. Viele Tensorflow Modelle bieten Textdateien an, die die Klassenlabels
     *         enthalten. Diese können als STring-Array konvertiert hier übergeben werden, um z.B.
     *         mittels der Ausgabemethode getOutputLabel() automatisch das Klassenlabel des Neurons
     *         auszugeben, auf das die Eingabe abgebildet wurde.
     */
    public void setOutputLabels(String[] pLabelArray) {
        labels = pLabelArray;        
    }

    /**
     * Diese Methode ist eine von zwei Möglichkeiten, das Ergebnis einer Klassifikation
     * zu erhalten nachdem ein Bild mit run() klassifiziert wurde. Diese Methode liefert
     * nur ein sinnvolles Ergebnis wenn zuvor mittels setOutputLabels(...) auch zum
     * Model passende Klassenlabels geladen wurden.
     * 
     * @return Zeichenkette, die die Klasse enthält, die laut Model-Vorhersage mit der
     * höchsten Wahrscheinlichkeit im Bild abgebildet ist.
     */
    public String getOutputLabel() {
        if (labels != null && output != null) {
            String erg = null;
            int maxPos = 0;
            float maxVal = -1;        
            for (int i=0; i<output.length; i++) {            
                if (output[i] > maxVal) {
                    maxVal = output[i];
                    maxPos = i;
                }
            }            
            return labels[maxPos];
        }
        else {
            return null;   
        }
    }
}